#!/usr/bin/env python

'''
Usage:
    {self} [--path=<path>] [-v -v...] <target> add <user>/<repo> [<name>] [--tracking=<branch>] [-a]
    {self} [--path=<path>] [-v -v...] <target> fork <user>/<repo> [<branch>] [--clone]
    {self} [--path=<path>] [-v -v...] <target> clone <user>/<repo> [<branch>]
    {self} [--path=<path>] [-v -v...] <target> create <user>/<repo> [--add]
    {self} [--path=<path>] [-v -v...] <target> delete <user>/<repo> [-f]
    {self} [--path=<path>] [-v -v...] <target> open [<user>/<repo>]
    {self} [--path=<path>] [-v -v...] <target> request [<user>/<repo>] (list|ls)
    {self} [--path=<path>] [-v -v...] <target> request [<user>/<repo>] fetch <request>
    {self} [--path=<path>] [-v -v...] <target> gist (list|ls) [<gist>]
    {self} [--path=<path>] [-v -v...] <target> gist clone <gist>
    {self} [--path=<path>] [-v -v...] <target> gist fetch <gist> [<gist_file>]
    {self} [--path=<path>] [-v -v...] <target> gist create [--secret] <description> [<gist_path> <gist_path>...]
    {self} [--path=<path>] [-v -v...] <target> gist delete <gist> [-f]
    {self} --help

Tool for managing remote repository services.

Commands:
    add                      Add the service as a remote on this repository
    clone                    Clones this repository from the service
    fork                     Fork (and clone) the repository from the service
    create                   Make this repository a new remote on the service
    delete                   Delete the remote repository
    gist                     Manages gist files
    request                  Handles requests for merge
    open                     Open the given or current repository in a browser

Options:
    <user>/<repo>            Repository to work with
    -p,--path=<path>         Path to work on [default: .]
    -v,--verbose             Makes it more chatty (repeat twice to see git commands)
    -h,--help                Shows this message

Options for add:
    <name>                   Name to use for the remote (defaults to name of repo)
    -t,--tracking <branch>   Makes this remote tracking for the current branch
    -a,--alone               Does not add the remote to the 'all' remote

Options for fork and clone:
    <branch>                 Branch to pull (when cloning) [default: master]
    --clone                  Clone locally after fork

Options for create:
    --add                    Add to local repository after creation

Options for delete:
    -f,--force               Do not ask for confirmation

Options for gist:
    <gist>                   Identifier of the gist to fetch
    <gist_file>              Name of the file to fetch
    <gist_path>              Name of the file or directory to use for a new gist.
                             If path is a directory, all files directly within it
                             will be pushed. If a list of path is given, all files
                             from them will be pushed.
    --secret                 Do not publicize gist when pushing

Configuration options:
    alias                    Name to use for the git remote
    url                      URL of the repository
    fqdn                     URL of the repository
    type                     Name of the service to use (github, gitlab, bitbucket)

Configuration example:

[gitrepo "gitlab"]
    token = yourapitoken
    alias = lab

[gitrepo "personal"]
    type = gitlab
    token = yourapitoken
    fqdn = custom.org

{self} version {version}, Copyright ⓒ2016 Bernard `Guyzmo` Pratz
{self} comes with ABSOLUTELY NO WARRANTY; for more informations
read the LICENSE file available in the sources, or check
out: http://www.gnu.org/licenses/gpl-2.0.txt
'''

from docopt import docopt

import os
import sys
import json
import logging
import pkg_resources

__version__ = pkg_resources.require('git-repo')[0].version
__author__ = 'Bernard `Guyzmo` Pratz <guyzmo+git_repo@m0g.net>'
__contributors__ = []

log_root = logging.getLogger()
log = logging.getLogger('git_repo')

if sys.version_info.major < 3: # pragma: no cover
    print('Please use with python version 3')
    sys.exit(1)

from .exceptions import ArgumentError
from .services.service import RepositoryService

from git import Repo, Git
from git.exc import InvalidGitRepositoryError, NoSuchPathError

def confirm(what, where):
    ans = input('Are you sure you want to delete the '
                '{} {} from the service?\n[yN]> '.format(what, where))
    if 'y' in ans:
        ans = input('Are you really sure? there\'s no coming back!\n'
                    '[type \'burn!\' to proceed]> ')
        if 'burn!' != ans:
            return False
    else:
        return False
    return True


def main(args):
    try:
        if args['--verbose'] >= 5:  # pragma: no cover
            print(args)
        if args['--verbose'] >= 4:  # pragma: no cover
            import http.client
            http.client.HTTPConnection.debuglevel = 1
            logging.getLogger("requests.packages.urllib3").setLevel(logging.DEBUG)
            logging.getLogger("requests.packages.urllib3").propagate = True
        if args['--verbose'] >= 3: # pragma: no cover
            Git.GIT_PYTHON_TRACE = 'full'
        if args['--verbose'] >= 2: # pragma: no cover
            Git.GIT_PYTHON_TRACE = True
            FORMAT = '> %(message)s'
            formatter = logging.Formatter(fmt=FORMAT)
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            logging.getLogger('git.cmd').removeHandler(logging.NullHandler())
            logging.getLogger('git.cmd').addHandler(handler)
        if args['--verbose'] >= 1: # pragma: no cover
            log_root.setLevel(logging.DEBUG)
        else: #pragma: no cover
            log_root.setLevel(logging.INFO)

        log.addHandler(logging.StreamHandler())

        # FIXME workaround for default value that is not correctly parsed in docopt
        if args['<branch>'] == None:
            args['<branch>'] = 'master'

        if 'GIT_WORK_TREE' in os.environ.keys() or 'GIT_DIR' in os.environ.keys(): #pragma: no cover
            del os.environ['GIT_WORK_TREE']

        if args['<user>/<repo>'] and '/' in args['<user>/<repo>']:
            if len(args['<user>/<repo>'].split('/')) > 2:
                raise ArgumentError('Too many slashes.'
                                    'Format of the parameter is <user>/<repo> or <repo>.')
            user, repo = args['<user>/<repo>'].split('/')
        else:
            user = None
            repo = args['<user>/<repo>']

        if args['create'] and not args['gist'] or args['add'] or args['delete'] and not args['gist'] or args['open'] or args['request']:
            # Try to resolve existing repository path
            try:
                try:
                    repository = Repo(os.path.join(args['--path'], repo))
                except NoSuchPathError:
                    repository = Repo(args['--path'])
            except InvalidGitRepositoryError:
                raise FileNotFoundError('Cannot find path to the repository.')
            service = RepositoryService.get_service(repository, args['<target>'])

            if args['create']:
                service.create(user, repo, add=args['--add'])
                log.info('Successfully created remote repository `{}`, '
                         'with local remote `{}`'.format(
                    service.format_path(repo, namespace=user),
                    service.name)
                )

            elif args['add']:
                service.add(repo, user,
                            name=args['<name>'],
                            tracking=args['--tracking'],
                            alone=args['--alone'])
                log.info('Successfully added `{}` as remote named `{}`'.format(
                    args['<user>/<repo>'],
                    service.name)
                )

            elif args['delete']:
                if not args['--force']: # pragma: no cover
                    if not confirm('repository', args['<user>/<repo>']):
                        return 0

                if user:
                    service.delete(repo, user)
                else:
                    service.delete(repo)
                log.info('Successfully deleted remote `{}` from {}'.format(
                    args['<user>/<repo>'],
                    service.name)
                )

            elif args['request']:
                if args['list'] or args['ls']:
                    log.info('List of open requests to merge:')
                    log.info(" {}\t{}\t{}".format('id', 'title'.ljust(60), 'URL'))
                    for pr in service.request_list(user, repo):
                        print("{}\t{}\t{}".format(pr[0].rjust(3), pr[1][:60].ljust(60), pr[2]))
                elif args['fetch'] and args['<request>']:
                    new_branch = service.request_fetch(user, repo, args['<request>'])
                    log.info('Successfully fetched request id `{}` of `{}` into `{}`!'.format(
                        args['<request>'],
                        args['<user>/<repo>'],
                        new_branch)
                    )

            elif args['open']:
                RepositoryService.get_service(None, args['<target>']).open(user, repo)

            return 0

        elif args['fork']:
            if not os.path.exists(repo):
                repo_path = os.path.join(args['--path'], repo)
                repository = Repo.init(repo_path)
                service = RepositoryService.get_service(repository, args['<target>'])
                service.fork(user, repo, branch=args['<branch>'], clone=args['--clone'])
                log.info('Successfully cloned repository {} in {}'.format(
                    args['<user>/<repo>'],
                    repo_path)
                )

                return 0
            else:
                raise FileExistsError('Cannot clone repository, '
                                      'a folder named {} already exists!'.format(repo))

        elif args['clone'] and not args['gist']:
            repo_path = os.path.join(args['--path'], repo)
            repository = Repo.init(repo_path)
            service = RepositoryService.get_service(repository, args['<target>'])
            service.clone(user, repo, args['<branch>'])
            log.info('Successfully cloned `{}` into `{}`!'.format(
                service.format_path(args['<user>/<repo>']),
                repo_path)
            )
            return 0

        elif args['gist']:
            service = RepositoryService.get_service(None, args['<target>'])
            service.connect()
            if args['list'] or args['ls']:
                if args['<gist>']:
                    log.info("{:15}\t{:>7}\t{}".format('language', 'size', 'name'))
                    for gist_file in service.gist_list(args['<gist>']):
                        print("{:15}\t{:7}\t{}".format(*gist_file))
                else:
                    log.info("{:56}\t{}".format('id', 'title'.ljust(60)))
                    for gist in service.gist_list():
                        print( "{:56}\t{}".format(gist[0], gist[1]))
            elif args['fetch']:
                # send gist to stdout, not using log.info on purpose here!
                print(service.gist_fetch(args['<gist>'], args['<gist_file>']))
            elif args['clone']:
                repo_path = os.path.join(args['--path'], args['<gist>'].split('/')[-1])
                service.repository = Repo.init(repo_path)
                service.gist_clone(args['<gist>'])
                log.info('Successfully cloned `{}` into `{}`!'.format( args['<gist>'], repo_path))
            elif args['create']:
                url = service.gist_create(args['<gist_path>'], args['<description>'], args['--secret'])
                log.info('Successfully created gist `{}`!'.format(url))
            elif args['delete']:
                if not args['--force']: # pragma: no cover
                    if not confirm('gist', args['<gist>']):
                        return 0

                service.gist_delete(args['<gist>'])
                log.info('Successfully deleted gist!')
            return 0

        log.error('Unknown action.')
        log.error('Please consult help page (--help).')
        return 1
    except Exception as err:
        log.error('Fatal error: {}'.format(err))
        if log_root.level == logging.DEBUG:
            log.exception('------------------------------------')
        return 2


def cli(): #pragma: no cover
    sys.exit(main(docopt(__doc__.format(self=sys.argv[0], version=__version__))))

if __name__ == '__main__': #pragma: no cover
    cli()



#!/usr/bin/env python

import logging
log = logging.getLogger('git_repo.github')

from ..service import register_target, RepositoryService, os
from ...exceptions import ResourceError, ResourceExistsError, ResourceNotFoundError

import github3

from git.exc import GitCommandError

@register_target('hub', 'github')
class GithubService(RepositoryService):
    fqdn = 'github.com'

    def __init__(self, *args, **kwarg):
        self.gh = github3.GitHub()
        super(GithubService, self).__init__(*args, **kwarg)

    def connect(self):
        try:
            self.gh.login(token=self._privatekey)
            self.username = self.gh.user()
        except github3.models.GitHubError as err:
            if err.code is 401:
                if not self._privatekey:
                    raise ConnectionError('Could not connect to Github. '
                                          'Please configure .gitconfig '
                                          'with your github private key.') from err
                else:
                    raise ConnectionError('Could not connect to Github. '
                                          'Check your configuration and try again.') from err

    def create(self, user, repo, add=False):
        try:
            self.gh.create_repo(repo)
        except github3.models.GitHubError as err:
            if err.code == 422 or err.message == 'name already exists on this account':
                raise ResourceExistsError("Project already exists.") from err
            else: # pragma: no cover
                raise ResourceError("Unhandled error.") from err
        if add:
            self.add(user=user, repo=repo, tracking=self.name)

    def fork(self, user, repo, branch='master', clone=False):
        log.info("Forking repository {}/{}…".format(user, repo))
        try:
            fork = self.gh.repository(user, repo).create_fork()
        except github3.models.GitHubError as err:
            if err.message == 'name already exists on this account':
                raise ResourceExistsError("Project already exists.") from err
            else: # pragma: no cover
                raise ResourceError("Unhandled error: {}".format(err)) from err
        self.add(user=user, repo=repo, name='upstream', alone=True)
        remote = self.add(repo=repo, user=self.username, tracking=self.name)
        if clone:
            self.pull(remote, branch)
        log.info("New forked repository available at {}/{}".format(self.url_ro,
                                                                   fork.full_name))

    def delete(self, repo, user=None):
        if not user:
            user = self.username
        try:
            repository = self.gh.repository(user, repo)
            if repository:
                result = repository.delete()
            if not repository or not result:
                raise ResourceNotFoundError('Cannot delete: repository {}/{} does not exists.'.format(user, repo))
        except github3.models.GitHubError as err: # pragma: no cover
            if err.code == 403:
                raise ResourcePermissionError('You don\'t have enough permissions for deleting the repository. '
                                              'Check the namespace or the private token\'s privileges') from err
            raise ResourceError('Unhandled exception: {}'.format(err)) from err

    def get_repository(self, user, repo):
        repository = self.gh.repository(user, repo)
        if not repository:
            raise ResourceNotFoundError('Cannot delete: repository {}/{} does not exists.'.format(user, repo))
        return repository

    def _format_gist(self, gist):
        return gist.split('https://gist.github.com/')[-1].split('.git')[0]

    def gist_list(self, gist=None):
        if not gist:
            for gist in self.gh.iter_gists(self.gh.user().name):
                yield (gist.html_url, gist.description)
        else:
            gist = self.gh.gist(self._format_gist(gist))
            if gist is None:
                raise ResourceNotFoundError('Gist does not exists.')
            for gist_file in gist.iter_files():
                yield (gist_file.language if gist_file.language else 'Raw text',
                        gist_file.size,
                        gist_file.filename)


    def gist_fetch(self, gist, fname=None):
        try:
            gist = self.gh.gist(self._format_gist(gist))
        except Exception as err:
            raise ResourceNotFoundError('Could not find gist') from err
        if gist.files == 1 and not fname:
            gist_file = list(gist.iter_files())[0]
        else:
            for gist_file in gist.iter_files():
                if gist_file.filename == fname:
                    break
            else:
                raise ResourceNotFoundError('Could not find file within gist.')

        return gist_file.content

    def gist_clone(self, gist):
        try:
            gist = self.gh.gist(gist.split('https://gist.github.com/')[-1])
        except Exception as err:
            raise ResourceNotFoundError('Could not find gist') from err
        remote = self.repository.create_remote('gist', gist.git_push_url)
        self.pull(remote, 'master')

    def gist_create(self, gist_pathes, description, secret=False):
        def load_file(fname, path='.'):
            with open(os.path.join(path, fname), 'r') as f:
                return {'content': f.read()}

        gist_files = dict()
        for gist_path in gist_pathes:
            if not os.path.isdir(gist_path):
                gist_files[os.path.basename(gist_path)] = load_file(gist_path)
            else:
                for gist_file in os.listdir(gist_path):
                    if not os.path.isdir(os.path.join(gist_path, gist_file)) and not gist_file.startswith('.'):
                        gist_files[gist_file] = load_file(gist_file, gist_path)

        gist = self.gh.create_gist(
                description=description,
                files=gist_files,
                public=not secret # isn't it obvious? ☺
            )

        return gist.html_url

    def gist_delete(self, gist_id):
        gist = self.gh.gist(self._format_gist(gist_id))
        if not gist:
            raise ResourceNotFoundError('Could not find gist')
        gist.delete()

    def request_list(self, user, repo):
        repository = self.gh.repository(user, repo)
        for pull in repository.iter_pulls():
            yield ( str(pull.number), pull.title, pull.links['issue'] )

    def request_fetch(self, user, repo, request, pull=False):
        if pull:
            raise NotImplementedError('Pull operation on requests for merge are not yet supported')
        try:
            for remote in self.repository.remotes:
                if remote.name == self.name:
                    local_branch_name = 'request/{}'.format(request)
                    self.fetch(
                        remote,
                        'pull/{}/head'.format(request),
                        local_branch_name
                    )
                    return local_branch_name
            else:
                raise ResourceNotFoundError('Could not find remote {}'.format(self.name))
        except GitCommandError as err:
            if 'Error when fetching: fatal: Couldn\'t find remote ref' in err.command[0]:
                raise ResourceNotFoundError('Could not find opened request #{}'.format(request)) from err
            raise err

    @property
    def user(self):
        return self.gh.user().name

